/////////////////////////////////////////////////// -*- mode:c++; -*- ////
//  Copyright (C) 2008 by Jonathan Marten				//
//  jjm@keelhaul.me.uk							//
//									//
//  This program is free software; you can redistribute it and/or	//
//  modify it under the terms of the GNU General Public License as	//
//  published by the Free Software Foundation; either version 2 of	//
//  the License, or (at your option) any later version.			//
//									//
//  This program is distributed in the hope that it will be useful,	//
//  but WITHOUT ANY WARRANTY; without even the implied warranty of	//
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the	//
//  GNU General Public License for more details.			//
//									//
//  You should have received a copy of the GNU General Public License	//
//  along with this program; if not, write to the			//
//  Free Software Foundation, Inc.,					//
//  59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.		//
//									//
//////////////////////////////////////////////////////////////////////////


//////////////////////////////////////////////////////////////////////////
//									//
//  This file is part of Kadoom, a KDE desktop front end to various	//
//  Doom game engines.  See http://www.keelhaul.me.uk/linux/kadoom	//
//  for the latest version and further information.			//
//									//
//									//
//  DisplayOptions							//
//  --------------							//
//									//
//  The generic 'Display' options dialogue, with common and useful	//
//  display options.  Some game engines may add options to or remove	//
//  them from this dialogue.						//
//									//
//////////////////////////////////////////////////////////////////////////

#include <kdebug.h>
#include <kdialog.h>
#include <klocale.h>
#include <kconfig.h>

#include <qsize.h>
#include <qcheckbox.h>
#include <qgroupbox.h>
#include <qlayout.h>
#include <qcombobox.h>
#include <qlabel.h>
#include <qspinbox.h>

#include "displayoptions.h"
#include "displayoptions.moc"

//////////////////////////////////////////////////////////////////////////
//									//
//  Preset display sizes						//
//									//
//////////////////////////////////////////////////////////////////////////

static QSize presetsizes[] =
{
	QSize(),					// custom
	QSize(320,256),
	QSize(640,480),
	QSize(800,600),
	QSize(1024,768),
	QSize(1280,1024)
};

//////////////////////////////////////////////////////////////////////////
//									//
//  Constructor/destructor						//
//									//
//////////////////////////////////////////////////////////////////////////

DisplayOptions::DisplayOptions(QWidget *parent,const QSize &size,DisplayOptions::Option opt)
	: KDialogBase(parent)
{
	setCaption(i18n("Display Options"));
	showButtonApply(false);				// uses "OK" and "cancel" only

	const int mh = KDialog::marginHint();		// conveniences, use them lots
	const int sh = KDialog::spacingHint();

	QWidget *b = new QWidget(this);			// central widget
	setMainWidget(b);
	QVBoxLayout *l1 = new QVBoxLayout(b,0,sh);	// main layout, vertical stack

	sizegroup = new QGroupBox(i18n("Size"),b);	// "Size" group box
	QGridLayout *l2 = new QGridLayout(sizegroup,4,7,mh,sh);	
							// its internal layout
	custsizex = new QSpinBox(320,2048,8,sizegroup);	// custom "Width" spin box
	l2->addWidget(custsizex,2,2,Qt::AlignLeft);
	QLabel *lab = new QLabel("",sizegroup);	// separator text label
	l2->addWidget(lab,2,3,Qt::AlignHCenter);
	custsizey = new QSpinBox(200,2048,8,sizegroup);	// custom "Height" spin box
	l2->addWidget(custsizey,2,4,Qt::AlignLeft);

	lab = new QLabel(i18n("Custom:"),sizegroup);	// their text label
	l2->addWidget(lab,2,1,Qt::AlignRight);
	lab->setBuddy(custsizex);

	sizecombo = new QComboBox(sizegroup);		// "Size" combo box
	sizecombo->insertItem(i18n("Custom..."));
	sizecombo->setCurrentItem(0);
	for (unsigned int i = 1; i<(sizeof(presetsizes)/sizeof(QSize)); ++i)
	{
		sizecombo->insertItem(QString("%1%2")
				      .arg(presetsizes[i].width())
				      .arg(presetsizes[i].height()));
		if (presetsizes[i]==size) slotChangedSize(i);
	}

	connect(sizecombo,SIGNAL(activated(int)),this,SLOT(slotChangedSize(int)));
	l2->addMultiCellWidget(sizecombo,1,1,2,5);

	lab = new QLabel(i18n("&Size:"),sizegroup);	// its text label
	l2->addWidget(lab,1,1,Qt::AlignRight);
	lab->setBuddy(sizecombo);

	l2->setRowSpacing(0,mh); l2->setRowSpacing(3,sh); l2->setRowStretch(3,9);
	l2->setColSpacing(0,sh); l2->setColSpacing(6,sh); l2->setColStretch(5,9);

	sizegroup->adjustSize();
	l1->addWidget(sizegroup);


	advgroup = new QGroupBox(i18n("Advanced"),b);	// "Advanced" group box
	l2 = new QGridLayout(advgroup,7,4,mh,sh);	// its internal layout

	expcombo = new QComboBox(advgroup);		// "Expansion" combo box
	expcombo->insertItem(i18n("1 (none)"));
	expcombo->insertItem("2");
	expcombo->insertItem("3");
	expcombo->setCurrentItem(0);
	l2->addWidget(expcombo,1,2);

	lab = new QLabel(i18n("Expansion:"),advgroup);	// its text label
	l2->addWidget(lab,1,1,Qt::AlignRight);
	lab->setBuddy(expcombo);

	l2->setRowSpacing(2,sh);

	noaccel = new QCheckBox(i18n("Disable accelerated graphics (only use if problems)"),advgroup);
	l2->addMultiCellWidget(noaccel,3,3,1,2,Qt::AlignLeft);

	nojoy = new QCheckBox(i18n("Do not use joystick"),advgroup);
	l2->addMultiCellWidget(nojoy,4,4,1,2,Qt::AlignLeft);

	nomouse = new QCheckBox(i18n("Do not use mouse"),advgroup);
	l2->addMultiCellWidget(nomouse,5,5,1,2,Qt::AlignLeft);

	l2->setRowSpacing(0,mh); l2->setRowSpacing(6,sh); l2->setRowStretch(6,9);
	l2->setColSpacing(0,sh); l2->setColSpacing(4,sh); l2->setColStretch(2,9);

	advgroup->adjustSize();
	l1->addWidget(advgroup);

	if (opt & DisplayOptions::NoAccel) noaccel->setChecked(true);
	if (opt & DisplayOptions::NoJoy) nojoy->setChecked(true);
	if (opt & DisplayOptions::NoMouse) nomouse->setChecked(true);

	if (opt & DisplayOptions::Expand2) expcombo->setCurrentItem(1);
	else if (opt & DisplayOptions::Expand3) expcombo->setCurrentItem(2);
}

//////////////////////////////////////////////////////////////////////////
//									//
//  Action slots							//
//									//
//////////////////////////////////////////////////////////////////////////

void DisplayOptions::slotChangedSize(int item)
{
	sizecombo->setCurrentItem(item);		// set for initialisation

	if (item>0)					// set to a numeric value
	{						// reflect value in spin boxes
		custsizex->setValue(presetsizes[item].width());
		custsizey->setValue(presetsizes[item].height());
	}

	custsizex->setEnabled(item==0);			// enable/disable spin boxes
	custsizey->setEnabled(item==0);
}

//////////////////////////////////////////////////////////////////////////
//									//
//  Returned results							//
//									//
//////////////////////////////////////////////////////////////////////////

DisplayOptions::Option DisplayOptions::options() const
{
	int opt = DisplayOptions::Null;
	if (noaccel->isChecked()) opt |= DisplayOptions::NoAccel;
	if (nojoy->isChecked()) opt |= DisplayOptions::NoJoy;
	if (nomouse->isChecked()) opt |= DisplayOptions::NoMouse;

	int exp = 1+expcombo->currentItem();
	if (exp==2) opt |= DisplayOptions::Expand2;
	else if (exp==3) opt |= DisplayOptions::Expand3;

	return ((DisplayOptions::Option) opt);
}


QSize DisplayOptions::size() const
{
	QSize sz = presetsizes[sizecombo->currentItem()];
	if (sz.isEmpty()) sz = QSize(custsizex->value(),custsizey->value());
	return (sz);
}
