/////////////////////////////////////////////////// -*- mode:c++; -*- ////
//  Copyright (C) 2008 by Jonathan Marten				//
//  jjm@keelhaul.me.uk							//
//									//
//  This program is free software; you can redistribute it and/or	//
//  modify it under the terms of the GNU General Public License as	//
//  published by the Free Software Foundation; either version 2 of	//
//  the License, or (at your option) any later version.			//
//									//
//  This program is distributed in the hope that it will be useful,	//
//  but WITHOUT ANY WARRANTY; without even the implied warranty of	//
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the	//
//  GNU General Public License for more details.			//
//									//
//  You should have received a copy of the GNU General Public License	//
//  along with this program; if not, write to the			//
//  Free Software Foundation, Inc.,					//
//  59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.		//
//									//
//////////////////////////////////////////////////////////////////////////


//////////////////////////////////////////////////////////////////////////
//									//
//  This file is part of Kadoom, a KDE desktop front end to various	//
//  Doom game engines.  See http://www.keelhaul.me.uk/linux/kadoom	//
//  for the latest version and further information.			//
//									//
//									//
//  ListParser								//
//  ----------								//
//									//
//  Simple sh(1)-like parser to split up a string (entered as the	//
//  "Advanced" engine options) into a list of command arguments.	//
//  Understands single and double quoting, and backslash escape.	//
//									//
//////////////////////////////////////////////////////////////////////////

#include <kdebug.h>

#include <qstring.h>
#include <qstringlist.h>
#include <qregexp.h>

#include <qdir.h>
#include <qfile.h>

#include "listparser.h"

//////////////////////////////////////////////////////////////////////////
//									//
//  Quote a list into a single string (not actually used in Kadoom).	//
//  Arguments containing whitespace are quoted appropriately, and	//
//  special characters are backslash escaped.				//
//									//
//////////////////////////////////////////////////////////////////////////

QString ListParser::quoteFromList(const QStringList &list)
{
	kdDebug(0) << k_funcinfo << "from list [" << list.join(",") << "]" << endl;

	QString e;
	const QRegExp needed("[\\s\"\']");		// things that need quoting

	for (QStringList::const_iterator it = list.begin(); it!=list.end(); ++it)
	{
		QString a = (*it);			// current item from list

		a.replace('\\',"\\\\");			// escape any backslashes
		if (a.contains(needed))			// now see if has space or quote
		{
			if (!a.contains('"'))		// can quote with double
			{
				a = '"'+a+'"';
			}
			else if (!a.contains('\''))	// can quote with single
			{
				a = '\''+a+'\'';
			}
			else				// unlucky, escape double quotes
			{				// and quote with that
				a.replace('"',"\\\"");
				a = '"'+a+'"';
			}
		}

		if (!e.isNull()) e += " ";
		e += a;					// append to result string
	}

	kdDebug() << "  quoted to [" << e << "]" << endl;
	return (e);
}

//////////////////////////////////////////////////////////////////////////
//									//
//  Parse a string into a list of arguments.  It is split up into	//
//  words separated by whitespace, taking account of quotes and		//
//  escaped characters.							//
//									//
//////////////////////////////////////////////////////////////////////////

enum ParserState					// string parsing state
{
	Space,						// white space between words
	Word,						// within an unquoted word
	DoubleQuote,					// within double-quoted string
	SingleQuote,					// within single-quoted string
	Escape						// after backslash escape
};

QStringList ListParser::parseToList(const QString &str)
{
	kdDebug() << k_funcinfo << "parsing [" << str << "]" << endl;

	QStringList list;
	const int len = str.length();			// length of input string
	if (len==0) return (list);			// quick check for empty

	QString a;					// current word so far
	ParserState state = Space;			// current parser state
	ParserState prev = state;			// previous state before escape

	for (int i = 0; i<len; ++i)			// scan the entire string
	{
		QChar ch = str.at(i);			// character at this position
		switch (state)
		{
case Space:		if (ch=='"')			// start of double quoted string
			{				// drop the starting quote
				a = QString::null;
				state = DoubleQuote;
			}
			else if (ch=='\'')		// start of single quoted string
			{				// drop the starting quote
				a = QString::null;
				state = SingleQuote;
			}
			else if (ch=='\\')		// escape next character
			{				// which also starts a word
				a = QString::null;
				prev = Word;
				state = Escape;
			}
			else if (!ch.isSpace())		// start of a new word
			{
				a = ch;
				state = Word;
			}
			break;

case Word:		if (ch.isSpace())		// end of the word
			{
				list += a;		// append this word to list
				a = QString::null;	// clear for next time
				state = Space;
			}
			else if (ch=='"')		// start of double quoted string
			{				// ignore the starting quote
				state = DoubleQuote;
			}
			else if (ch=='\'')		// start of single quoted string
			{				// ignore the starting quote
				state = SingleQuote;
			}
			else if (ch=='\\')		// escape next character
			{
				prev = state;
				state = Escape;
			}
			else a += ch;			// append to this word
			break;

case DoubleQuote:	if (ch=='"')			// end of quoted string
			{
				state = Word;		// check for whitespace next
			}
			else if (ch=='\\')		// escape next character
			{
				prev = state;
				state = Escape;
			}
			else a += ch;			// append to this word
			break;

case SingleQuote:	if (ch=='\'')			// end of quoted string
			{
				state = Word;		// check for whitespace next
			}
			else if (ch=='\\')		// escape next character
			{
				prev = state;
				state = Escape;
			}
			else a += ch;			// append to this word
			break;

case Escape:		a += ch;			// append literal character
			state = prev;			// return to previous state
			break;
		}
	}
	if (!a.isNull()) list += a;			// last word at end of string

	kdDebug(0) << k_funcinfo << "  parsed to [" << list.join(",") << "]" << endl;
	return (list);
}
