/////////////////////////////////////////////////// -*- mode:c++; -*- ////
//  Copyright (C) 2008 by Jonathan Marten				//
//  jjm@keelhaul.me.uk							//
//									//
//  This program is free software; you can redistribute it and/or	//
//  modify it under the terms of the GNU General Public License as	//
//  published by the Free Software Foundation; either version 2 of	//
//  the License, or (at your option) any later version.			//
//									//
//  This program is distributed in the hope that it will be useful,	//
//  but WITHOUT ANY WARRANTY; without even the implied warranty of	//
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the	//
//  GNU General Public License for more details.			//
//									//
//  You should have received a copy of the GNU General Public License	//
//  along with this program; if not, write to the			//
//  Free Software Foundation, Inc.,					//
//  59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.		//
//									//
//////////////////////////////////////////////////////////////////////////


//////////////////////////////////////////////////////////////////////////
//									//
//  This file is part of Kadoom, a KDE desktop front end to various	//
//  Doom game engines.  See http://www.keelhaul.me.uk/linux/kadoom	//
//  for the latest version and further information.			//
//									//
//									//
//  KadoomMainWindow							//
//  ----------------							//
//									//
//  The application main window, menu bar and buttons.  Most of what	//
//  appears is the appropriate game engine widget (a subclass of	//
//  GenericGame).  The entire main window, including that, is		//
//  regenerated (by KadoomMainObject) when the game engine changes.	//
//									//
//////////////////////////////////////////////////////////////////////////

#include <kdebug.h>
#include <kstandarddirs.h>
#include <kglobal.h>
#include <klocale.h>
#include <kstatusbar.h>
#include <kprocess.h>
#include <kmessagebox.h>
#include <kconfig.h>
#include <kiconloader.h>
#include <kaction.h>
#include <kstdaction.h>
#include <kapplication.h>

#include <qpushbutton.h>
#include <qlayout.h>
#include <qlabel.h>
#include <qframe.h>
#include <qwidgetstack.h>
#include <qcursor.h>

#include "gamemanager.h"
#include "genericgame.h"
#include "setupwizard.h"
#include "remotefiles.h"

#include "mainwindow.h"
#include "mainwindow.moc"

//////////////////////////////////////////////////////////////////////////
//									//
//  Command line options						//
//									//
//////////////////////////////////////////////////////////////////////////

extern bool optionTestingOnly;				// from "main"

//////////////////////////////////////////////////////////////////////////
//									//
//  Constructor/destructor						//
//									//
//////////////////////////////////////////////////////////////////////////

KadoomMainWindow::KadoomMainWindow()
	: KMainWindow(NULL,"")
{
	kdDebug(0) << k_funcinfo << endl;

	QWidget *vb = new QWidget(this);		// the main widget and layout
	QGridLayout *l1 = new QGridLayout(vb,3,1,KDialog::marginHint(),KDialog::spacingHint());

	gamewidget = NULL;				// none created yet
	setAcceptDrops(false);				// not accepted just anywhere

	readProperties(KGlobal::config());		// read our global options
							// now we know engine to use, so
	gameframe = new QWidgetStack(vb);		// create game engine widget
	createEngine(gameframe);
	l1->addWidget(gameframe,0,0);
	l1->setRowStretch(0,1);
	l1->setRowSpacing(1,KDialog::marginHint());
							// the "Start" button below
	startbutton = new QPushButton(KGlobal::iconLoader()->loadIcon("kadoom_play",KIcon::Small),i18n("Start Play"),vb);
	startbutton->setDefault(true);
	startbutton->adjustSize();
	startbutton->setSizePolicy(QSizePolicy::Fixed,QSizePolicy::Fixed);
	l1->addWidget(startbutton,2,0,Qt::AlignCenter);

	setCentralWidget(vb);				// set that as main widget
	setupActions();					// menu and toolbar actions

	//KGlobal::dirs()->addResourceDir("data",QDir::current().path());
	setupGUI(KMainWindow::Keys+KMainWindow::Save+KMainWindow::Create,"kadoomui.rc");

	gamewidget->setupActions(this);			// add game engine's actions
	gamewidget->setWadDataDirectory(gamedata);	// tell it this setting

	connect(gamewidget,SIGNAL(readyToPlay(bool,bool)),this,SLOT(slotReadyToPlay(bool,bool)));
	connect(startbutton,SIGNAL(clicked()),this,SLOT(slotStartPlay()));
							// now can accept signals
	gamewidget->checkForNewSaves();			// update list of saved games
	gamewidget->checkPlayAllowed();			// update enable states
}


KadoomMainWindow::~KadoomMainWindow()
{
	gamewidget->saveOptions(KGlobal::config());	// save global options
	delete gamewidget;
	delete gameframe;

	kdDebug(0) << k_funcinfo << "done" << endl;
}

//////////////////////////////////////////////////////////////////////////
//									//
//  Menu and button actions						//
//									//
//////////////////////////////////////////////////////////////////////////

void KadoomMainWindow::setupActions()
{
	KStdAction::quit(kapp, SLOT(quit()), actionCollection());

	KAction *prefAction = KStdAction::preferences(this, SLOT(optionsPreferences()), actionCollection());
	prefAction->setEnabled(false);			// not currently implemented

	playgameaction = new KAction(startbutton->text(),"kadoom_play",CTRL+Key_S,
				     this, SLOT(slotStartPlay()),
				     actionCollection(),"start_play");

	new KAction("Choose Game Engine...", "options",0,
		    this,SLOT(slotConfigureEngine()),
		    actionCollection(),"settings_gameplayer");

	new KAction("Clear Download Cache","kadoom_clearcache",0,
		    this,SLOT(slotClearCache()),
		    actionCollection(),"clear_cache");
}

//////////////////////////////////////////////////////////////////////////
//									//
//  Saving and loading global options, including the game engine type	//
//									//
//////////////////////////////////////////////////////////////////////////

void KadoomMainWindow::saveProperties(KConfig *config)
{
	config->setGroup("Player");
	config->writeEntry("gametype",gametype);
	config->writePathEntry("gameexec",gameexec);
	config->writePathEntry("gamedata",gamedata);
}


void KadoomMainWindow::readProperties(KConfig *config)
{
	config->setGroup("Player");
	gametype = config->readEntry("gametype");
	gameexec = config->readPathEntry("gameexec");
	gamedata = config->readPathEntry("gamedata");
}

//////////////////////////////////////////////////////////////////////////
//									//
//  Action slots							//
//									//
//////////////////////////////////////////////////////////////////////////

void KadoomMainWindow::optionsPreferences()
{							// there aren't any, yet
}


void KadoomMainWindow::slotClearCache()
{
	QApplication::setOverrideCursor(QCursor(Qt::WaitCursor));
	RemoteFiles::self()->clearCache();
	QApplication::restoreOverrideCursor();
}


void KadoomMainWindow::slotConfigureEngine()
{
	SetupWizard d(i18n("Choose Game Engine"),this);
	if (d.exec())
	{
		kdDebug(0) << k_funcinfo << "wizard finished" << endl;
		kdDebug(0) << "  key=" << d.getGameKey() << endl;
		kdDebug(0) << "  name=" << d.getGameName() << endl;
		kdDebug(0) << "  exec=" << d.getExecutable() << endl;
		kdDebug(0) << "  data=" << d.getDataDir() << endl;

		gametype = d.getGameKey();
		gameexec = d.getExecutable();
		gamedata = d.getDataDir();
		saveProperties(KGlobal::config());

		emit engineChanged();
	}
}

//////////////////////////////////////////////////////////////////////////
//									//
//  Game engine widget							//
//									//
//////////////////////////////////////////////////////////////////////////

void KadoomMainWindow::createEngine(QWidget *parent)
{
	kdDebug(0) << k_funcinfo << "type=" << gametype << endl;

	if (gametype.isNull())				// first time, not yet set
	{
		slotConfigureEngine();			// start the wizard now
		if (gametype.isNull())			// cancelled that, abort now
		{
			KMessageBox::sorry(NULL,i18n("No game engine selected."));
			exit(1);
		}
	}

	if (gamewidget!=NULL)				// there is one already
	{
		gamewidget->saveOptions(KGlobal::config());
		gameframe->removeWidget(gamewidget);
		delete gamewidget;
	}
							// create widget for engine
	gamewidget = GameManager::self()->create(gametype,parent);
	if (gamewidget==NULL)
	{
		KMessageBox::error(NULL,i18n("<qt>No support available for engine type '<b>%1</b>'").arg(gametype),"Fatal error");
		exit(1);
	}

	setCaption(gamewidget->gameInfo()->name);
	gameframe->addWidget(gamewidget);
	gamewidget->readOptions(KGlobal::config());
}

//////////////////////////////////////////////////////////////////////////
//									//
//  Running the engine							//
//									//
//////////////////////////////////////////////////////////////////////////

void KadoomMainWindow::slotReadyToPlay(bool ok,bool demo)
{
	startbutton->setEnabled(ok);
	startbutton->setText(demo ? i18n("Start Demo") : i18n("Start Play"));
	playgameaction->setEnabled(ok);
}


void KadoomMainWindow::slotStartPlay()
{
	QStringList cmd(gameexec);			// command and argument list

	const QString err = gamewidget->getPlayCommand(cmd);
	if (!err.isNull())				// build and validate command
	{						// no message if just cancelled
		if (err!="!") KMessageBox::error(this,i18n("<qt>Cannot run the game; check the settings.<p>%1").
                                                 arg(err),QString::null,KMessageBox::Notify|KMessageBox::AllowLink);
		return;
	}

	kdDebug(0) << k_funcinfo << endl << "cmd = [" << cmd.join(" ") << "]" << endl;
	if (optionTestingOnly) return;			// option "-n" on command line

	KProcess *proc = new KProcess(this);		// new process to run game
	*proc << cmd;					// set command and arguments
	connect(proc,SIGNAL(processExited(KProcess *)),this,SLOT(slotFinishedPlay(KProcess *)));
							// go, tell me when it finishes
	if (!proc->start(KProcess::NotifyOnExit) || !proc->isRunning())
	{
		KMessageBox::error(this,i18n("<qt>Cannot start the game; check that the command '<b>%1</b>'<br>is installed and available.<p>You may need to set PATH to find it.").arg(cmd[0]));
		return;
	}

	gamewidget->setEnabled(false);			// no interaction while running
}


void KadoomMainWindow::slotFinishedPlay(KProcess *proc)
{
	QString reason = QString::null;
	if (proc->normalExit())
	{
		if (proc->exitStatus()!=0) reason = i18n("exited with status %1").arg(proc->exitStatus());
	}
	else
	{
		if (proc->signalled()) reason = i18n("terminated by signal %1").arg(proc->exitSignal());
		else reason = i18n("finished unexpectedly");
	}

	kdDebug(0) << k_funcinfo << "reason=[" << reason << "]" << endl;
	if (!reason.isNull())
	{
		KMessageBox::error(this,i18n("Problem running game: %1").arg(reason));
	}

	delete proc; proc = NULL;			// finished with the process

	gamewidget->checkForNewSaves();			// locate any new saved games
	gamewidget->setEnabled(true);			// now can interact again
}
