/////////////////////////////////////////////////// -*- mode:c++; -*- ////
//  Copyright (C) 2008 by Jonathan Marten				//
//  jjm@keelhaul.me.uk							//
//									//
//  This program is free software; you can redistribute it and/or	//
//  modify it under the terms of the GNU General Public License as	//
//  published by the Free Software Foundation; either version 2 of	//
//  the License, or (at your option) any later version.			//
//									//
//  This program is distributed in the hope that it will be useful,	//
//  but WITHOUT ANY WARRANTY; without even the implied warranty of	//
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the	//
//  GNU General Public License for more details.			//
//									//
//  You should have received a copy of the GNU General Public License	//
//  along with this program; if not, write to the			//
//  Free Software Foundation, Inc.,					//
//  59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.		//
//									//
//////////////////////////////////////////////////////////////////////////


//////////////////////////////////////////////////////////////////////////
//									//
//  This file is part of Kadoom, a KDE desktop front end to various	//
//  Doom game engines.  See http://www.keelhaul.me.uk/linux/kadoom	//
//  for the latest version and further information.			//
//									//
//									//
//  MatchResolver							//
//  -------------							//
//									//
//  Resolve a match for a source URL from a list of possibilities.	//
//  This may be a local or remote directory listing, or the list of	//
//  entries from an archive.						//
//									//
//////////////////////////////////////////////////////////////////////////

#include <kdebug.h>
#include <kurl.h>
#include <kapplication.h>
#include <klocale.h>

#include <qstringlist.h>

#include "matchrequester.h"
#include "remotefiles.h"

#include "matchresolver.h"

//////////////////////////////////////////////////////////////////////////
//									//
//  Constructor/destructor						//
//									//
//  Will search for a match for the base filename of 'src' with the	//
//  extension 'ext'.  If the source has a reference (after "#"), then	//
//  this will override the file name.					//
//									//
//////////////////////////////////////////////////////////////////////////

MatchResolver::MatchResolver(const KURL &src,const QString &ext)
{
	kdDebug(0) << k_funcinfo << "for " << src << " with ext " << ext << endl;

	expectBase = RemoteFiles::nameBase(src.hasRef() ? src.ref() : src.path());
	expectExt = ext.upper();			// remember these for search
	source = src.fileName();			// remember this for dialogue

	foundEntry = QString::null;			// nothing found yet
}

//////////////////////////////////////////////////////////////////////////
//									//
//  Perform the match							//
//									//
//////////////////////////////////////////////////////////////////////////

MatchResolver::Status MatchResolver::match(const QStringList &list,const QString &message)
{
	kdDebug(0) << k_funcinfo << "expect " << expectBase << " with ext " << expectExt << endl;

	matchState = Nothing;

	QStringList extMatches;
	for (QStringList::const_iterator it = list.begin(); it!=list.end(); ++it)
	{
		QString entry = (*it);
		//kdDebug(0) << "  entry " << entry << endl;

		if (matchState==Nothing)		// nothing found yet
		{					// note the first entry
			foundEntry = entry;
			matchState = First;
			//kdDebug(0) << "  match state 1 with " << foundEntry << endl;
		}

		QString entryBase = RemoteFiles::nameBase(entry);
		QString entryExt = RemoteFiles::nameExt(entry);
		//kdDebug(0) << "  check " << entryBase << " with ext " << entryExt << endl;

		if (entryBase==expectBase && entryExt==expectExt)
		{					// exact match
			foundEntry = entry;
			matchState = NameExact;
			//kdDebug(0) << "  match state 4 with " << foundEntry << endl;
			break;				// no need to look further
		}

		if (entryBase.lower()==expectBase.lower() && entryExt==expectExt && matchState<NameNear)
		{					// case independent name match
			foundEntry = entry;
			matchState = NameNear;
			//kdDebug(0) << "  match state 3 with " << foundEntry << endl;
		}

		if (entryExt==expectExt && matchState<NameNear)
		{					// extension match
			extMatches += entry;
			if (matchState<Extension) matchState = Extension;
			//kdDebug(0) << "  match state 2 added " << entry << endl;
		}
	}

	if (matchState<Extension)
	{
		kdDebug(0) << "  nothing appropriate found" << endl;
		return (matchState);
	}

	if (matchState==Extension)
	{
		if (extMatches.count()==1) foundEntry = extMatches.first();
		else
		{
			MatchRequester d(kapp->mainWidget(),message,extMatches);
			d.show();			// use dialogue, abort if cancelled
			if (!d.exec()) return (Cancelled);

			foundEntry = d.selectedEntry();
			matchState = (d.rememberChoice() ? Remember : UserSelect);
		}
	}

	kdDebug(0) << k_funcinfo << "result " << matchState << " found " << foundEntry << endl;
	return (matchState);				// result of match
}

//////////////////////////////////////////////////////////////////////////
//									//
//  Returned results							//
//									//
//////////////////////////////////////////////////////////////////////////

QString MatchResolver::found() const
{
	return (foundEntry);
}
