/////////////////////////////////////////////////// -*- mode:c++; -*- ////
//  Copyright (C) 2008 by Jonathan Marten				//
//  jjm@keelhaul.me.uk							//
//									//
//  This program is free software; you can redistribute it and/or	//
//  modify it under the terms of the GNU General Public License as	//
//  published by the Free Software Foundation; either version 2 of	//
//  the License, or (at your option) any later version.			//
//									//
//  This program is distributed in the hope that it will be useful,	//
//  but WITHOUT ANY WARRANTY; without even the implied warranty of	//
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the	//
//  GNU General Public License for more details.			//
//									//
//  You should have received a copy of the GNU General Public License	//
//  along with this program; if not, write to the			//
//  Free Software Foundation, Inc.,					//
//  59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.		//
//									//
//////////////////////////////////////////////////////////////////////////


//////////////////////////////////////////////////////////////////////////
//									//
//  This file is part of Kadoom, a KDE desktop front end to various	//
//  Doom game engines.  See http://www.keelhaul.me.uk/linux/kadoom	//
//  for the latest version and further information.			//
//									//
//									//
//  RemoteFiles								//
//  -----------								//
//									//
//  This is the heart of Kadoom's main user convenience:  downloading,	//
//  caching and extracting of remote and archive files as if they were	//
//  local.  The same access is actually used for all files, being a	//
//  no-op for suitable local ones.  There is only one unique instance.	//
//									//
//////////////////////////////////////////////////////////////////////////

#include <kdebug.h>
#include <kurl.h>
#include <kapplication.h>
#include <kstandarddirs.h>
#include <kmdcodec.h>
#include <kmimetype.h>
#include <kzip.h>
#include <klocale.h>
#include <kprotocolinfo.h>

#include <kio/netaccess.h>

#include <qstringlist.h>
#include <qdir.h>
#include <qfile.h>

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#ifdef HAVE_NANOSLEEP
#include <time.h>
#else
#include <unistd.h>
#endif

#include <string.h>
#ifdef HAVE_ERRNO_H
#include <errno.h>
#endif

#include "matchresolver.h"

#include "remotefiles.h"

//////////////////////////////////////////////////////////////////////////
//									//
//  Instance and access							//
//									//
//////////////////////////////////////////////////////////////////////////

static RemoteFiles *instance = NULL;


RemoteFiles *RemoteFiles::self()
{
	if (instance==NULL) instance = new RemoteFiles();
	return (instance);
}

//////////////////////////////////////////////////////////////////////////
//									//
//  Constructor/destructor						//
//									//
//////////////////////////////////////////////////////////////////////////

RemoteFiles::RemoteFiles()
{
	QString cachePath = KGlobal::dirs()->saveLocation("cache",(PACKAGE "/"),true);
	cacheDir = QDir(cachePath);

	kdDebug(0) << k_funcinfo << "cache at " << cacheDir.path() << endl;
}

//////////////////////////////////////////////////////////////////////////
//									//
//  File name parsing							//
//									//
//////////////////////////////////////////////////////////////////////////

QString RemoteFiles::nameBase(const QString &path)
{
	int i = path.findRev('/');
	QString name = (i>0 ? path.mid(i+1) : path);
	i = name.findRev('.');
	return (i>0 ? name.left(i) : name);
}


QString RemoteFiles::nameExt(const QString &path)
{
	int i = path.findRev('/');
	QString name = (i>0 ? path.mid(i+1) : path);
	i = name.findRev('.');
	return (i>0 ? name.mid(i+1).upper() : QString::null);
}

//////////////////////////////////////////////////////////////////////////
//									//
//  File tests								//
//									//
//////////////////////////////////////////////////////////////////////////

QString RemoteFiles::isAccessibleFile(const KURL &url)
{
	if (url.isLocalFile()) return (url.path());
	KURL localURL = KIO::NetAccess::mostLocalURL(url,kapp->mainWidget());
	kdDebug(0) << k_funcinfo << "for " << url.url() << " local " << localURL.url() << " result " << (localURL!=url) << endl;
	return (localURL!=url ? localURL.path() : QString::null);
}


KURL RemoteFiles::isArchiveFile(const KURL &url)
{
	KURL u = url;

	if (!KProtocolInfo::determineMimetypeFromExtension(u.protocol()))
	{						// need to download to check?
		if (!u.isLocalFile())			// yes, really need to download?
		{
			QString localFile = downloadIntoCache(u,nameExt(localFile));
			if (localFile.isNull()) return (KURL());
			u = KURL(localFile);
		}
	}
							// local file here, now can ask
	QString mimetype = KMimeType::findByURL(u)->name();
	kdDebug(0) << k_funcinfo << "for " << u.prettyURL() << " mimetype = " << mimetype << endl;

	return ((mimetype=="application/x-zip") ? u : KURL());
}

//////////////////////////////////////////////////////////////////////////
//									//
//  File caching							//
//									//
//////////////////////////////////////////////////////////////////////////

QString RemoteFiles::cacheFile(const KURL &src,const QString &ext)
{
	KURL u = src;
	u.setRef(QString::null);

	KMD5 hash(u.url().local8Bit());
	const QString cacheFile = cacheDir.filePath(QString(hash.hexDigest())+"."+ext);
	kdDebug(0) << k_funcinfo << "for " << u.prettyURL() << " ext " << ext << " = " << cacheFile << endl;
	return (cacheFile);
}


static void sleepfor(long int usecs)
{
#ifdef HAVE_NANOSLEEP
	struct timespec t;
	t.tv_sec = (time_t) 0;
	t.tv_nsec = 1000*usecs;
	nanosleep(&t,NULL);				// POSIX
#else
#ifdef HAVE_USLEEP
	usleep(usecs);					// BSD/XOpen
#else
	sleep(1);					// the best we can do!
#endif
#endif
}


QString RemoteFiles::downloadIntoCache(const KURL &src,const QString &ext)
{
	QString cachedFile = cacheFile(src,ext);
	if (!QFile::exists(cachedFile))
	{
		if (!KIO::NetAccess::download(src,cachedFile,kapp->mainWidget()))
		{					// KIO detected error
			lastErr = KIO::NetAccess::lastErrorString();
			if (!lastErr.isEmpty())	return (QString::null);
		}					// if no message, continue
							// will test if exists later
		if (!QFile::exists(cachedFile))
		{
			//  Sometimes, trying QFile::exists() immediately after the download
			//  doesn't return TRUE even though the file is there.  Wait for a
			//  very short time and try again.
			kdDebug(0) << k_funcinfo << "downloaded " << src.prettyURL() << " as " << cachedFile << " but didn't exist!" << endl;

			//kdDebug() << "sleeping to try again 1..." << endl;
			sleepfor(1000);
			if (!QFile::exists(cachedFile))
			{
				//kdDebug() << "sleeping to try again 2..." << endl;
				sleepfor(10000);
				if (!QFile::exists(cachedFile))
				{
					lastErr = i18n("<qt>Failed to download <b>%1</b>").arg(src.prettyURL());
					return (QString::null);
				}
			}
		}

		kdDebug(0) << k_funcinfo << "downloaded " << src.prettyURL() << " to " << cachedFile << endl;
	}

	return (cachedFile);
}


const QString RemoteFiles::getWithoutExtracting(const KURL &src)
{
	//kdDebug(0) << k_funcinfo << "for " << src.prettyURL() << endl;
	lastErr = QString::null;			// no error yet

	QString localFile = isAccessibleFile(src);	// see if a local file, if so
	if (localFile.isNull())				// no need to download
	{						// otherwise download it
		localFile = downloadIntoCache(src,nameExt(src.path()));
		if (localFile.isNull()) return (QString::null);
	}

	return (localFile);
}

//////////////////////////////////////////////////////////////////////////
//									//
//  Match caching							//
//									//
//////////////////////////////////////////////////////////////////////////

void RemoteFiles::saveMatch(const KURL &src,const QString &ext,const QString &match)
{
	const QString matchFile = cacheFile(src,ext+"match");
	kdDebug(0) << k_funcinfo << "for " << src.prettyURL() << " ext " << ext << " to " << matchFile << endl;

	QFile f(matchFile);
	if (!f.open(IO_WriteOnly))
	{
		kdDebug(0) << "  match file " << matchFile << " write error!" << endl;
		return;
	}

	QTextStream t(&f);
	t << match << '\n';				// save match to file
	f.close();					// finished with cache file
}


QString RemoteFiles::cachedMatch(const KURL &src,const QString &ext)
{
	if (src.hasRef()) return (QString::null);	// will use reference later

	const QString matchFile = cacheFile(src,ext+"match");
	kdDebug(0) << k_funcinfo << "for " << src.prettyURL() << " ext " << ext << " from " << matchFile << endl;

	QFile f(matchFile);
	if (!f.exists()) return (QString::null);	// check for cache file first
	if (!f.open(IO_ReadOnly))			// exists, but problem?
	{
		kdDebug(0) << "  match file " << matchFile << " read error!" << endl;
		return (QString::null);
	}

	QTextStream t(&f);
	QString match = t.readLine();			// read match from file
	f.close();					// finished with cache file
	return (match);
}

//////////////////////////////////////////////////////////////////////////
//									//
//  File access								//
//									//
//////////////////////////////////////////////////////////////////////////

const QString RemoteFiles::get(const KURL &src,const char *ext)
{
	kdDebug(0) << k_funcinfo << "for " << src.prettyURL() << " ext " << (ext!=NULL ? ext : "*") << endl;

	QString localFile = getWithoutExtracting(src);
	if (localFile.isNull()) return (QString::null);

	QString mimetype = KMimeType::findByPath(localFile)->name();
	//kdDebug(0) << "  mimetype = " << mimetype << endl;
	if (mimetype=="application/x-zip")
	{
		QString expectBase = nameBase(src.hasRef() ? src.ref() : src.path());
		QString expectExt = (ext!=NULL ? QString(ext).upper() : QString::null);
		//kdDebug(0) << "  expect " << expectBase << " with ext " << expectExt << endl;

		QString zipEntry = cachedMatch(src,expectExt);
		if (zipEntry.isNull())			// see if already cached
		{					// if not, scan the archive
			KZip zipFile(localFile);
			if (!zipFile.open(IO_ReadOnly))
			{
				lastErr = i18n("<qt>Error reading Zip file <b>%2</b>, %1").arg(strerror(errno),localFile);
				return (QString::null);
			}

			const KArchiveDirectory *zipDir = zipFile.directory();
			const QStringList zipEntries = zipDir->entries();
			zipFile.close();		// don't need this any more

			MatchResolver res(src,expectExt);
			MatchResolver::Status matchState = res.match(zipEntries,
								     i18n("<qt>Select entry from archive file <b>%1</b>:").arg(src.fileName()));
			if (matchState==MatchResolver::Cancelled)
			{
				lastErr = "!";		// this means "cancelled"
				return (QString::null);
			}

			if (matchState<MatchResolver::Extension)
			{
				if (src.isLocalFile())
				{
					lastErr = i18n("<qt>No suitable %1 member found in Zip file<br><a href=\"%3\">%2</a>")
					               .arg(expectExt,localFile,KURL::fromPathOrURL(localFile).url());
				}
				else
				{
					lastErr = i18n("<qt>No suitable %1 member found in downloaded Zip file<br><a href=\"%3\">%2</a>")
					               .arg(expectExt,src.prettyURL(),src.url());
				}
				return (QString::null);
			}

			zipEntry = res.found();
			if (matchState==MatchResolver::Remember && !src.hasRef())
			{
				saveMatch(src,expectExt,zipEntry);
			}
		}

		QString srcFile = QString("zip:")+localFile+"/"+zipEntry;
		//kdDebug(0) << "  src in zip = " << srcFile << endl;

		localFile = downloadIntoCache(srcFile,expectExt);
		if (localFile.isNull()) return (QString::null);
	}

	kdDebug(0) << k_funcinfo << "returning " << localFile << endl;
	return (localFile);
}

//////////////////////////////////////////////////////////////////////////
//									//
//  Cache clearing							//
//									//
//////////////////////////////////////////////////////////////////////////

void RemoteFiles::clearCache()
{
	QDir d(cacheDir);
	d.setFilter(QDir::Files);			// all files in cache directory
	QStringList cacheContents = d.entryList();
	for (QStringList::const_iterator it = cacheContents.begin();
	     it!=cacheContents.end(); ++it)
	{
		QFile::remove(d.absFilePath(*it));	// remove each file
	}
}
